/*
 * deviced
 *
 * Copyright (c) 2014 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <stdbool.h>
#include <hw/touchscreen.h>
#include <libsyscommon/dbus-system.h>
#include <vconf.h>
#include <dlfcn.h>
#include "core/devices.h"
#include "core/common.h"
#include "core/log.h"
#include "battery/power-supply.h"
#include "display/core.h"
#include "display/display-ops.h"

enum ps_mode {
	POWERSAVING_OFF,
	POWERSAVING_ON,
};

static struct touchscreen_device *touchscreen_dev;
static int touchscreen_enable = DEVICE_OPS_STATUS_START;
static int powersaving_support = true;

static int touchscreen_start(enum device_flags flags);
static int touchscreen_stop(enum device_flags flags);

static struct display_config *_display_conf;
static struct _backlight_ops *_backlight_ops;

static void touchscreen_wakeup_status(keynode_t *key, void *data)
{
	if (!key)
		return;

	if (!_display_conf || !_backlight_ops)
		return;

	_display_conf->touch_wakeup = vconf_keynode_get_bool(key);

	if (_backlight_ops->get_lcd_power() != DPMS_ON) {
		if (_display_conf->touch_wakeup)
			touchscreen_start(NORMAL_MODE);
		else
			touchscreen_stop(TOUCH_SCREEN_OFF_MODE);
	}
}

static int touchscreen_powersaving(enum ps_mode mode)
{
	bool state;
	int ret;

	if (!powersaving_support)
		return -ENOTSUP;

	if (touchscreen_enable != DEVICE_OPS_STATUS_START)
		return 0;

	if (mode == POWERSAVING_ON)
		state = true;
	else if (mode == POWERSAVING_OFF)
		state = false;
	else
		return -EINVAL;

	if (touchscreen_dev && touchscreen_dev->set_powersaving) {
		ret = touchscreen_dev->set_powersaving(state);
		if (ret < 0) {
			_E("Failed to set touch powersaving node.");
			return -EIO;
		}
	}

	_I("Set touch powersaving %s.", state ? "ON" : "OFF");

	return 0;
}
static int touchscreen_probe(void *data)
{
	int ret;
	struct hw_info *info;

	if (touchscreen_dev)
		return 0;

	ret = hw_get_info(TOUCHSCREEN_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (ret < 0) {
		_E("Failed to load touchscreen shared library: %d", ret);
		return -ENODEV;
	}

	if (!info->open) {
		_E("Failed to open touchscreen device: open(NULL)");
		return -EPERM;
	}

	ret = info->open(info, NULL, (struct hw_common **)&touchscreen_dev);
	if (ret < 0) {
		_E("Failed to get touchscreen device structure: %d", ret);
		return -EPERM;
	}

	_I("Touchscreen device structure load success.");
	return 0;
}

static void touchscreen_exit(void *data)
{
	struct hw_info *info;

	if (!touchscreen_dev)
		return;

	info = touchscreen_dev->common.info;
	assert(info);

	info->close((struct hw_common *)touchscreen_dev);
	touchscreen_dev = NULL;
}

static int touchscreen_set_state(enum touchscreen_state state)
{
	int ret;
	char *act;

	if (!touchscreen_dev) {
		_E("Touchscreen device structure is not loaded.");
		return -ENOENT;
	}

	if (state != TOUCHSCREEN_ON && state != TOUCHSCREEN_OFF) {
		_E("Invalid parameter.");
		return -EINVAL;
	}

	if (!touchscreen_dev->set_state) {
		_E("Touchscreen state change is not supported.");
		return -ENOTSUP;
	}

	act = (state == TOUCHSCREEN_ON) ? "enable" : "disable";

	ret = touchscreen_dev->set_state(state);
	if (ret == 0)
		_I("Success to %s touchscreen.", act);
	else
		_E("Failed to %s touchscreen: %d", act, ret);

	return ret;
}

static int touchscreen_execute(void *data)
{
	touchscreen_enable = (int)((intptr_t)data);
	return 0;
}

static int touchscreen_start(enum device_flags flags)
{
	int state, ret;

	if (touchscreen_enable != DEVICE_OPS_STATUS_START)
		return 0;

	if (!_backlight_ops || !_backlight_ops->get_lcd_power)
		return -ENOTSUP;

	/*
	 * It is safe to turn touchscreen on before powersaving on/off.
	 */
	ret = touchscreen_set_state(TOUCHSCREEN_ON);

	state = _backlight_ops->get_lcd_power();

	if (state == DPMS_OFF)
		touchscreen_powersaving(POWERSAVING_ON);
	else
		touchscreen_powersaving(POWERSAVING_OFF);

	return ret;
}

static int touchscreen_stop(enum device_flags flags)
{
	if (!_display_conf || !_backlight_ops) {
		_I("Touchscreen is not initialized.");
		goto exit;
	}

	/*
	 * Touchscreen control priority
	 *  1. Touchscreen module control by execute(touchscreen_enable)
	 *  2. Device charger status by charging dock
	 *  3. Touch wakeup setting(touch_wakeup_enable) value
	 *  4. Normal mode(Basic request)
	 */
	if (touchscreen_enable != DEVICE_OPS_STATUS_START)
		return 0;

	if (flags & TOUCH_SCREEN_OFF_MODE) {
		touchscreen_powersaving(POWERSAVING_OFF);
		return touchscreen_set_state(TOUCHSCREEN_OFF);
	}

	if (battery.online_type >= CHARGER_TYPE_WIRE) {
		_I("Battery charging one %d %d.", battery.online, battery.online_type);
		return touchscreen_powersaving(POWERSAVING_ON);
	}

	if (_display_conf->touch_wakeup) {
		_I("Touch wakeup enabled.");
		return touchscreen_powersaving(POWERSAVING_ON);
	}

exit:
	touchscreen_powersaving(POWERSAVING_OFF);
	return touchscreen_set_state(TOUCHSCREEN_OFF);
}

static int touchscreen_dump(FILE *fp, int mode, void *dump_data)
{
	int ret;
	enum touchscreen_state state;

	if (!touchscreen_dev)
		return 0;

	ret = touchscreen_dev->get_state(&state);
	if (ret < 0)
		_E("Failed to get touchscreen state: %d", ret);

	if (state == TOUCHSCREEN_ON)
		LOG_DUMP(fp, "Touchscreen is enabled\n");
	else
		LOG_DUMP(fp, "Touchscreen is disabled\n");

	return 0;
}

static GVariant *dbus_touchscreen_enable(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;
	ret = touchscreen_start(NORMAL_MODE);

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_touchscreen_disable(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;
	ret = touchscreen_stop(TOUCH_SCREEN_OFF_MODE);

	return g_variant_new("(i)", ret);
}

static const dbus_method_s dbus_methods[] = {
	{ "Enable",     NULL,    "i",  dbus_touchscreen_enable },
	{ "Disable",    NULL,    "i",  dbus_touchscreen_disable },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_TOUCH,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static void touchscreen_init(void *data)
{
	int ret, val;

	_display_conf = dlsym(disp_plgn.handle, "display_conf");
	if (!_display_conf)
		_E("Failed to obtain address of display_conf.");

	_backlight_ops = dlsym(disp_plgn.handle, "backlight_ops");
	if (!_backlight_ops)
		_E("Failed to obtain address of backlight_ops.");

	if (touchscreen_dev && touchscreen_dev->set_powersaving) {
		ret = touchscreen_dev->set_powersaving(0);
		if (ret < 0)
			powersaving_support = false;
		else
			powersaving_support = true;
	} else {
		powersaving_support = false;
	}

	vconf_notify_key_changed(VCONFKEY_SETAPPL_TOUCH_WAKEUP_ENABLE,
			touchscreen_wakeup_status, NULL);
	ret = vconf_get_bool(VCONFKEY_SETAPPL_TOUCH_WAKEUP_ENABLE, &val);
	if (_display_conf && ret == 0)
		_display_conf->touch_wakeup = val;


	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_TOUCH, &dbus_interface);

	if (ret < 0)
		_E("Failed to init dbus method. (%d)", ret);
}

static const struct device_ops touchscreen_device_ops = {
	DECLARE_NAME_LEN("touchscreen"),
	.probe    = touchscreen_probe,
	.init     = touchscreen_init,
	.exit     = touchscreen_exit,
	.start    = touchscreen_start,
	.stop     = touchscreen_stop,
	.dump     = touchscreen_dump,
	.execute  = touchscreen_execute,
};

DEVICE_OPS_REGISTER(&touchscreen_device_ops)

