/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdbool.h>

#include <vconf.h>
#include <hw/usb_gadget.h>
#include <libsyscommon/dbus-system.h>

#include "core/log.h"

#include "usb.h"
#include "usb-debug.h"

/* Legacy signals */
#define SIGNAL_STATE_CHANGED  "StateChanged"
#define SIGNAL_MODE_CHANGED   "ModeChanged"
#define SIGNAL_CONFIG_ENABLED "ConfigEnabled"
#define CHANGE_USB_MODE       "ChangeUsbMode"

static int get_usb_state(void)
{
	int ret, val;
	int result = 0;

	ret = vconf_get_int(VCONFKEY_SYSMAN_USB_STATUS, &val);
	if (ret != VCONF_OK) {
		_E("Failed to get vconf value for USB status: %d", vconf_get_ext_errno());
		return ret;
	}

	if (val == VCONFKEY_SYSMAN_USB_DISCONNECTED)
		result = val;
	else {
		result = VCONFKEY_SYSMAN_USB_CONNECTED;
		if (val == VCONFKEY_SYSMAN_USB_AVAILABLE)
			result |= VCONFKEY_SYSMAN_USB_AVAILABLE;
	}

	return result;
}

void broadcast_usb_state_changed(void)
{
	int ret;
	unsigned int state;
	static unsigned int prev_state = UINT_MAX;

	state = get_usb_state();
	if (state == prev_state)
		return;
	prev_state = state;

	_I("USB state(%u) changed.", state);

	ret = dbus_handle_emit_dbus_signal(NULL,
			DEVICED_PATH_USB,
			DEVICED_INTERFACE_USB,
			SIGNAL_STATE_CHANGED,
			g_variant_new("(u)", state));
	if (ret < 0)
		_E("Failed to send USB dbus signal.");
}

static void change_usb_client_mode(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)

{
	int ret;
	unsigned int mode;
	int req_vconf = -1;

	g_variant_get(param, "(i)", &req_vconf);
	if (req_vconf < 0) {
		_E("Failed to get USB req_vconf.");
		return;
	}

	_I("USB mode is changed by dbus signal to %d.", req_vconf);

	mode = get_mode_bitmap_from_vconf(req_vconf);

	if (mode == USB_FUNCTION_INVALID) {
		_E("Failed to convert vconf to USB mode. There is no mode matches up with vconf %d.", req_vconf);
		return;
	}

	/* Deviced must never run in USB_FUNCTION_NONE mode. */
	if (mode == USB_FUNCTION_NONE) {
		_E("Ignore this request. There is USB_FUNCTION_NONE mode matches up with vconf %d.", req_vconf);
		return;
	}

	ret = usb_change_mode(mode);
	if (ret < 0)
		_E("Failed to change USB mode: %d", ret);
}

/* dbus methods */
static GVariant *get_usb_client_state(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	unsigned int state;

	state = get_usb_state();

	return g_variant_new("(u)", state);
}

static GVariant *get_usb_client_mode(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	unsigned int mode;

	mode = usb_state_get_current_mode();

	return g_variant_new("(u)", mode);
}

static const dbus_method_s dbus_methods[] = {
	{ "GetState", NULL, "u", get_usb_client_state },  /* from Tizen 2.3 */
	{ "GetMode", NULL, "u", get_usb_client_mode  },   /* from Tizen 2.3 */
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_USB,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};


int usb_dbus_init(void)
{
	int ret;

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_USB, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method for USB: %d", ret);

	ret = subscribe_dbus_signal(NULL, DEVICED_PATH_USB,
		DEVICED_INTERFACE_USB, "ChangeUsbMode",
		change_usb_client_mode, NULL, NULL);
	if (ret <= 0) {
		_E("Failed to subscribe dbus signal for USB: %d", ret);
		return ret;
	}

	return 0;
}
