/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <vconf.h>
#include <bundle.h>
#include <eventsystem.h>
#include <hw/usb_gadget.h>

#include "core/log.h"
#include "apps/apps.h"
#include "extcon/extcon.h"

#include "usb.h"
#include "usb-debug.h"

static int noti_id = -1;

static unsigned int usb_current_mode = USB_FUNCTION_NONE;
static unsigned int usb_selected_mode = USB_FUNCTION_NONE;

static extcon_usb_state_e usb_connection = USB_DISCONNECTED;

/**
 * Important
 * You must keep the order for the specific items.
 *
 * There may be several mode_v for a mode.
 * In this case, the representative value should be placed on top.
 *
 * In case of SET_USB_SDB_DIAG and SET_USB_DIAG_SDB, the SET_USB_SDB_DIAG should be above.
 * As another example, SET_USB_RNDIS_SDB should be at the top of them.
 */
static const struct _usb_mode_mapping_table {
	int          mode_v; /* Integer defined by vconf */
	unsigned int mode;   /* Bitmap of usb function combination */
} usb_mode_mapping_table[] = {
	/* Hack for performance. In most cases this is the value. */
	{SET_USB_DEFAULT,         USB_FUNCTION_MTP | USB_FUNCTION_ACM},
	{SET_USB_SDB,             USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB},

	{SET_USB_NONE,            USB_FUNCTION_NONE},
	{SET_USB_RNDIS,           USB_FUNCTION_RNDIS},
	{SET_USB_RNDIS_DIAG,      USB_FUNCTION_RNDIS | USB_FUNCTION_DIAG},
	{SET_USB_DIAG_RMNET,      USB_FUNCTION_DIAG | USB_FUNCTION_RMNET},
	{SET_USB_ACM_SDB_DM,      USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_DM},

	{SET_USB_SDB_DIAG,        USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_DIAG},
	{SET_USB_DIAG_SDB,        USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_DIAG},

	{SET_USB_RNDIS_SDB,       USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS},
	{SET_USB_RNDIS_TETHERING, USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS},
	{SET_USB_RNDIS_SDB_ACM,   USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS},
};

unsigned int get_mode_bitmap_from_vconf(int mode_v)
{
	int i;
	int array_len = sizeof(usb_mode_mapping_table)/sizeof(usb_mode_mapping_table[0]);

	for (i = 0; i < array_len; i++) {
		if (usb_mode_mapping_table[i].mode_v == mode_v)
			return usb_mode_mapping_table[i].mode;
	}

	return USB_FUNCTION_INVALID;
}

static int get_mode_vconf_from_bitmap(unsigned int mode)
{
	int i;
	int array_len = sizeof(usb_mode_mapping_table)/sizeof(usb_mode_mapping_table[0]);

	/* Caution: The order to search the usb_mode_mapping_table must start with a low index. */
	for (i = 0; i < array_len; i++) {
		if (usb_mode_mapping_table[i].mode == mode)
			return usb_mode_mapping_table[i].mode_v;
	}

	return SET_USB_INVALID;
}

static void usb_state_send_system_event(int status)
{
	bundle *b;
	const char *str;

	switch (status) {
	case VCONFKEY_SYSMAN_USB_DISCONNECTED:
		str = EVT_VAL_USB_DISCONNECTED;
		break;

	case VCONFKEY_SYSMAN_USB_CONNECTED:
		str = EVT_VAL_USB_CONNECTED;
		break;

	case VCONFKEY_SYSMAN_USB_AVAILABLE:
		str = EVT_VAL_USB_AVAILABLE;
		break;

	default:
		return;
	}

	_I("USB system_event (%s)", str);

	b = bundle_create();
	bundle_add_str(b, EVT_KEY_USB_STATUS, str);
	eventsystem_send_system_event(SYS_EVENT_USB_STATUS, b);
	bundle_free(b);
}

void usb_state_retrieve_selected_mode(void)
{
	int mode_v;
	unsigned int mode;
	const unsigned int default_mode = USB_FUNCTION_MTP | USB_FUNCTION_ACM;

	/* Never return here because deviced must never run in USB_FUNCTION_NONE mode */
	if (vconf_get_int(VCONFKEY_USB_SEL_MODE, &mode_v) != VCONF_OK) {
		mode_v = SET_USB_INVALID;
		_E("Failed to get vconf value for USB sel mode: %d", vconf_get_ext_errno());
	}

	mode = get_mode_bitmap_from_vconf(mode_v);

	/*
	 * Deviced must never run in USB_FUNCTION_NONE mode.
	 * So if vconf value is invalid, deviced uses the default usb mode internally.
	 * Keep the problematic vconf values in order to define the problem correctly.
	 */
	switch (mode) {
	case USB_FUNCTION_INVALID: /* Invalid vconf */
		usb_selected_mode = default_mode;
		_E("Failed to convert vconf to USB mode. There is no mode matches up with vconf %d. Use default mode=0x%x.", mode_v, default_mode);
		break;

	case USB_FUNCTION_NONE: /* Invalid vconf */
		usb_selected_mode = default_mode;
		_E("There is USB_FUNCTION_NONE USB mode matches up with vconf %d. Use default mode=0x%x.", mode_v, default_mode);
		break;

	default: /* Success */
		usb_selected_mode = mode;
		_I("Succeeded to convert vconf to USB mode. vconf=%d, mode=0x%x.", mode_v, mode);
		break;
	}

	/*
	 * Tizen has no way of having different vconf value in engineer mode and user mode.
	 * So, alternatively, always enable SDB in engineer mode.
	 *
	 * Expected default sel_mode value in vconf in user     mode: 1 (MTP + ACM)
	 * Expected default sel_mode value in vconf in engineer mode: 2 (MTP + ACM + SDB)
	 */
#ifdef ENGINEER_MODE
	_I("Engineer mode. USB mode 0x%x, debug state %d", usb_selected_mode, get_usb_debug_state());

	if (!(usb_selected_mode & USB_FUNCTION_SDB)) {
		usb_selected_mode = USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB;
		(void)usb_state_set_selected_mode(usb_selected_mode);
	}
#endif

	/* To sync with vconf for debug mode */
	set_usb_debug_state((bool)(usb_selected_mode & USB_FUNCTION_SDB));
}

/* Cache of vconf_get_int(VCONFKEY_USB_SEL_MODE) */
unsigned int usb_state_get_selected_mode(void)
{
	return usb_selected_mode;
}

/* Since it is changed externally by dbus and vconf, it should be processed even if it has the same value as before. */
int usb_state_set_selected_mode(unsigned int mode)
{
	int ret;
	int mode_v;

	usb_selected_mode = mode;

	mode_v = get_mode_vconf_from_bitmap(mode);
	if (mode_v == SET_USB_INVALID) {
		_E("Failed to convert USB selected_mode to vconf. There is no vconf matches up with USB mode 0x%x", mode);
		return -EINVAL;
	}

	ret = vconf_set_int(VCONFKEY_USB_SEL_MODE, mode_v);
	if (ret != VCONF_OK)
		_E("Failed to set vconf value for USB selected mode: %d", vconf_get_ext_errno());

	return ret;
}

/* Cache of vconf_get_int(VCONFKEY_USB_CUR_MODE) */
unsigned int usb_state_get_current_mode(void)
{
	return usb_current_mode;
}

/* Because it is only changed by deviced, it is only processed when the vconf actually changes. */
int usb_state_set_current_mode(unsigned int mode)
{
	int ret = 0;
	int mode_v;
	static int old_mode_v = -1;

	usb_current_mode = mode;

	mode_v = get_mode_vconf_from_bitmap(mode);
	if (mode_v == SET_USB_INVALID) {
		_E("Failed to convert USB current_mode to vconf. There is no vconf matches up with mode 0x%x", mode);
		return -EINVAL;
	}

	if (old_mode_v != mode_v) {
		old_mode_v = mode_v;
		ret = vconf_set_int(VCONFKEY_USB_CUR_MODE, mode_v);
		if (ret < 0)
			_E("Failed to set vconf value for USB current mode: %d", vconf_get_ext_errno());
	}

	return ret;
}

extcon_usb_state_e usb_state_get_connection(void)
{
	return usb_connection;
}

void usb_state_set_connection(extcon_usb_state_e conn)
{
	usb_connection = conn;
}

static void media_noti_cb(GVariant *var, void *user_data, GError *err)
{
	int id = 0;

	if (!var) {
		if (err)
			_E("USB media notification error: %s", err->message);
		return;
	}

	if (!dh_get_param_from_var(var, "(i)", &id)) {
		_E("Failed to get variant(%s): no USB notification message", g_variant_get_type_string(var));
		goto out;
	}

	noti_id = id;
	_D("USB media notification message(%d)", noti_id);

out:
	g_variant_unref(var);
}

static void add_notification_handler(void)
{
	int ret;

	if (noti_id < 0) {
		ret = add_async_notification("MediaDeviceNotiOn", media_noti_cb, NULL, NULL);
		if (ret < 0)
			_E("Failed to add USB notification for usb connection: %d", ret);
	}
}

static void remove_notification_handler(void)
{
	int ret;

	if (noti_id >= 0) {
		ret = remove_notification("MediaDeviceNotiOff", noti_id);
		if (ret < 0)
			_E("Failed to remove USB notification for usb connection: %d", ret);
		else
			noti_id = -1;
	}
}

void change_usb_state_notification_handler(unsigned int mode)
{
	if (mode & USB_FUNCTION_MTP)
		add_notification_handler();
	else if (mode == USB_FUNCTION_NONE)
		remove_notification_handler();
}

/*
 * USB_DISCONNECTED                  : VCONFKEY_SYSMAN_USB_DISCONNECTED
 * USB_CONNECTED + USB_FUNCTION_NONE : VCONFKEY_SYSMAN_USB_CONNECTED
 * USB_CONNECTED + USB_FUNCTION_OOOO : VCONFKEY_SYSMAN_USB_AVAILABLE
 *
 * When connecting the usb cable     : "disconnected"(previous value) -> "connected" -> "available"
 * When disconnecting the usb cable  : "available"(previous value) -> "dosconnected"
 * When changing usb mode            : "available"(previous value) -> "connected" -> "available"
 *
 * When USB cable is connected but USB initialization fails : It stays "connected" without going to "available"
 *
 */
void send_usb_state_changed_event(int status)
{
	static int old_status = -1;

	if ((status != VCONFKEY_SYSMAN_USB_CONNECTED) &&
		(status != VCONFKEY_SYSMAN_USB_AVAILABLE) &&
		(status != VCONFKEY_SYSMAN_USB_DISCONNECTED)) {
		_E("Invalid USB state changed event (%d)", status);
		return;
	}

	if (old_status == status)
		return;

	old_status = status;

	usb_state_send_system_event(status);

	if (vconf_set_int(VCONFKEY_SYSMAN_USB_STATUS, status) != VCONF_OK)
		_E("Failed to set vconf value for USB status: %d", vconf_get_ext_errno());

	/* Caution: calls after vconf_set_int(VCONFKEY_SYSMAN_USB_STATUS) */
	broadcast_usb_state_changed();
}

