/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdbool.h>

#include <vconf.h>
#include <hw/usb_gadget.h>

#include "core/log.h"
#include "core/device-notifier.h"

#include "usb.h"

static bool tethering_state;

static void usb_tethering_changed(keynode_t *key, void *data)
{
	int mode;
	bool curr;

	if (!key)
		return;

	mode = vconf_keynode_get_int(key);
	curr = mode & VCONFKEY_MOBILE_HOTSPOT_MODE_USB;

	if (curr == tethering_state) {
		_I("USB tethering mode is already %s.", curr ? "ON" : "OFF");
		return;
	}

	tethering_state = curr;

	_I("USB tethering mode is changed to %s.", curr ? "ON" : "OFF");

	device_notify(DEVICE_NOTIFIER_USB_TETHERING_MODE, (void *)curr);
}

static int usb_tethering_mode_changed(void *on)
{
	int ret;
	unsigned new_mode;
	static unsigned int saved_prev_mode = USB_FUNCTION_NONE;
	const unsigned int curr_mode = usb_state_get_selected_mode();
	const unsigned int predefined_rndis_mode_off = USB_FUNCTION_ACM | USB_FUNCTION_MTP;
	const unsigned int predefined_rndis_mode_on = USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS;

	/*
	 * On: Use predefined mode (USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS)
	 * Off: Use predefined mode (USB_FUNCTION_MTP | USB_FUNCTION_ACM) + USB_FUNCTION_SDB(conditional)
	 *
	 * Caution: If usb debug menu was enabled, add USB_FUNCTION_SDB to new mode to use.
	 *
	 * When tethering is enabled, additionally enable usb debug(USB_FUNCTION_SDB) unconditionally.
	 * Since usb debug is always enabled when you turn off usb tethering, there is no way to know if usb debug was enabled or not.
	 * So before enabling usb tethering, save the previous state.
	 * This saved value is used to check if the debug menu was enabled or not when you turn off usb tethering.
	 *
	 * In addition, if someone changes the usb mode while tethering on,
	 * use the latest usb mode to determine the usb debug mode instead of using the saved value.
	 *
	 */

	if ((bool)on) {
		new_mode = predefined_rndis_mode_on;
		saved_prev_mode = curr_mode;
	} else {
		new_mode = predefined_rndis_mode_off;

		if (curr_mode != predefined_rndis_mode_on) /* someone changes the usb mode while tethering on, use latest usb mode */
			saved_prev_mode = curr_mode;

		if(saved_prev_mode & USB_FUNCTION_SDB)
			new_mode |= USB_FUNCTION_SDB;

		saved_prev_mode = USB_FUNCTION_NONE;
	}

	ret = usb_change_mode(new_mode);
	if (ret < 0)
		_E("Failed to change USB mode to (0x%x).", new_mode);

	return ret;
}

void add_usb_tethering_handler(void)
{
	if (vconf_notify_key_changed(VCONFKEY_MOBILE_HOTSPOT_MODE, usb_tethering_changed, NULL) != VCONF_OK)
		_E("Failed to add USB tethering handler.");

	register_notifier(DEVICE_NOTIFIER_USB_TETHERING_MODE, usb_tethering_mode_changed);
}

void remove_usb_tethering_handler(void)
{
	vconf_ignore_key_changed(VCONFKEY_MOBILE_HOTSPOT_MODE, usb_tethering_changed);

	unregister_notifier(DEVICE_NOTIFIER_USB_TETHERING_MODE, usb_tethering_mode_changed);
}
